from bdb import BdbQuit
import datetime
import functools
from pathlib import Path
from pprint import pformat
import sys
import time
import traceback
import types
import typing
from typing import Any, Callable, Dict, Mapping, Optional, Union

import click
from loguru import logger
import yaml

from covid_shared import paths


class YamlIOMixin:
    """Mixin class for reading and writing data from yaml files."""

    @staticmethod
    def _load(in_file: typing.TextIO) -> Dict:
        return yaml.load(in_file)

    @staticmethod
    def _write(data: Dict[str, Any], out_file: typing.TextIO):
        yaml.dump(data, out_file)


class Metadata:
    """Base metadata class.  Looks and feels like a dict with a limited API.

    This class is meant for recording metadata for applications and run
    environments and forwarding that information across pipeline stages.

    This class provides the interface for file I/O, but does not implement
    it.

    """

    def __init__(self, *args, **kwargs):
        self._metadata = {}

    def update(self, metadata_update: Mapping):
        """Dictionary style update of metadata."""
        # Inefficient, by centralizes error handling.
        for key, value in metadata_update.items():
            self[key] = value

    def update_from_file(self, metadata_key: str, metadata_file: typing.TextIO):
        """Loads a metadata file from disk and stores it in the key."""
        logger.warning('Base metadata information cannot be constructed from a file. Returning empty metadata.')

    def to_dict(self):
        """Give back a dict version of the metadata."""
        return self._metadata.copy()

    def __getitem__(self, metadata_key: str):
        return self._metadata[metadata_key]

    def __setitem__(self, metadata_key: str, value: Any):
        if metadata_key in self:
            # This feels like a weird use of KeyError.  AttributeError also
            # feels wrong.  Maybe write a custom error later.
            raise KeyError(f'Metadata key {metadata_key} has already been set.')
        self._metadata[metadata_key] = value

    def __contains__(self, metadata_key: str):
        return metadata_key in self._metadata

    def dump(self, metadata_file: typing.TextIO):
        """Interface to dump metadata to disk."""
        logger.warning('Base metadata class should not be used to dump information to a file.')
        pass

    def __repr__(self):
        return self.__class__.__name__


class RunMetadata(Metadata, YamlIOMixin):
    """Metadata class meant specifically for application runners.

    Silently records profiling and provenance information.

    """

    def __init__(self, *args, **kwargs):
        self._start = time.time()
        super().__init__(*args, **kwargs)
        self['start_time'] = datetime.datetime.now().strftime("%Y_%m_%d_%H_%M_%S")

    def update_from_path(self, metadata_key: str, metadata_path: Union[str, Path]):
        """Updates metadata from a metadata file path."""
        metadata_path = Path(metadata_path)
        if not metadata_path.name == paths.METADATA_FILE_NAME.name:
            raise ValueError('Can only update from `metadata.yaml` files.')
        with metadata_path.open() as metadata_file:
            self.update_from_file(metadata_key, metadata_file)

    def update_from_file(self, metadata_key: str, metadata_file: typing.TextIO):
        """Loads a metadata file from disk and stores it in the key."""
        self._metadata[metadata_key] = yaml.full_load(metadata_file)

    def dump(self, metadata_file_path: Union[str, Path]):
        self._metadata['run_time'] = f"{time.time() - self._start:.2f} seconds"
        try:
            with Path(metadata_file_path).open('w') as metadata_file:
                self._write(self._metadata, metadata_file)
        except FileNotFoundError:
            logger.warning(f'Output directory for {metadata_file.name} does not exist. Dumping metadata to console.')
            click.echo(pformat(self._metadata))


def monitor_application(func: types.FunctionType, logger_: Any, with_debugger: bool,
                        app_metadata: Optional[Metadata] = None) -> Callable:
    """Monitors an application for errors and injects a metadata container.

    Catches records them if they occur. Can also be configured to drop
    a user into an interactive debugger on failure.

    Parameters
    ----------
    func
        The application function to monitor.
    logger_
        The application logger
    with_debugger
        Whether the monitor drops a user into a pdb session on application
        failure.
    app_metadata
        Record for application metadata.

    """
    if app_metadata is None:
        app_metadata = Metadata()
    @functools.wraps(func)
    def _wrapped(*args, **kwargs):
        result = None
        try:
            # Record arguments for the run and inject the metadata
            app_metadata['main_function'] = f'{func.__module__}:{func.__name__}'
            app_metadata['run_arguments'] = get_function_full_argument_mapping(func, app_metadata,
                                                                               *args, **kwargs)
            result = func(app_metadata, *args, **kwargs)
            app_metadata['success'] = True
        except (BdbQuit, KeyboardInterrupt):
            app_metadata['success'] = False
            app_metadata['error_info'] = 'User interrupt.'
        except Exception as e:
            # For general errors, write exception info to the metadata.
            app_metadata['success'] = False
            logger_.exception("Uncaught exception {}".format(e))
            exc_type, exc_value, exc_traceback = sys.exc_info()
            app_metadata['error_info'] = {
                'exception_type': exc_type,
                'exception_value': exc_value,
                'exc_traceback': traceback.format_tb(exc_traceback)
            }
            if with_debugger:
                import pdb
                traceback.print_exc()
                pdb.post_mortem()
        finally:
            return app_metadata, result
    return _wrapped


def update_with_previous_metadata(run_metadata: RunMetadata, input_root: Path) -> RunMetadata:
    """Convenience function for updating metadata from an input source."""
    key = str(input_root.resolve()).replace(' ', '_').replace('-', '_').lower() + '_metadata'
    with (input_root / paths.METADATA_FILE_NAME).open() as input_metadata_file:
        run_metadata.update_from_file(key, input_metadata_file)
    return run_metadata


def get_function_full_argument_mapping(func: types.FunctionType, *args, **kwargs) -> Dict:
    """Get a dict representation of all args and kwargs for a function."""
    # Grab all variables in the enclosing namespace.  Args will be first.
    # Note: This may rely on the CPython implementation.  Not sure.
    arg_names = func.__code__.co_varnames
    arg_vals = [str(arg) for arg in args]
    # Zip ignores extra items in the second arg.  Use that property to catch
    # all positional args and ignore kwargs.
    run_args = dict(zip(arg_names, arg_vals))
    run_args.update({k: str(v) for k, v in kwargs.items()})
    return run_args



from datetime import datetime
from pathlib import Path
import random
from typing import Callable

import pytest

from covid_shared import cli_tools, paths

MOCK_DATETIME = datetime(2020, 4, 25, 17, 5, 55)
ABSOLUTE_PATH_1 = Path('/absolute/path/one')
ABSOLUTE_PATH_2 = Path('/absolute/path/two')
RELATIVE_PATH = Path('some/relative/path')


@pytest.fixture
def mock_datetime(mocker):

    class _mydatetime:

        @classmethod
        def now(cls):
            return MOCK_DATETIME

        @classmethod
        def strptime(cls, s, _format):
            return datetime.strptime(s, _format)

    return mocker.patch('covid_shared.cli_tools.run_directory.datetime.datetime', _mydatetime)


@pytest.fixture()
def empty_run_dir_root(tmp_path: Path):
    cli_tools.setup_directory_structure(tmp_path, with_production=True)
    return tmp_path


@pytest.fixture()
def run_dir_root(empty_run_dir_root, mock_datetime):
    for _ in range(10):
        run_dir = cli_tools.get_run_directory(empty_run_dir_root)
        run_dir.mkdir()
    return empty_run_dir_root


def _get_random_run_dir(root_dir: Path):
    # Run dirs are dot separated.
    sub_dirs = [d for d in root_dir.iterdir() if '.' in d.name]
    return random.choice(sub_dirs)


def test_get_run_directory(tmp_path: Path, mock_datetime):
    dir_date = MOCK_DATETIME.strftime('%Y_%m_%d')
    run_dir = cli_tools.get_run_directory(tmp_path)
    assert run_dir.name == f'{dir_date}.01'
    run_dir.mkdir()

    run_dir = cli_tools.get_run_directory(tmp_path)
    assert run_dir.name == f'{dir_date}.02'


def test_setup_directory_structure_no_production(tmp_path: Path):
    cli_tools.setup_directory_structure(tmp_path)
    assert tmp_path.exists()
    for subdir in [paths.BEST_LINK, paths.LATEST_LINK]:
        assert (tmp_path / subdir).exists()
        assert (tmp_path / subdir).is_dir()
        assert not (tmp_path / subdir).is_symlink()

    assert not (tmp_path / paths.PRODUCTION_RUN).exists()


def test_setup_directory_structure(tmp_path: Path):
    cli_tools.setup_directory_structure(tmp_path, with_production=True)
    assert tmp_path.exists()
    for subdir in [paths.BEST_LINK, paths.LATEST_LINK, paths.PRODUCTION_RUN]:
        assert (tmp_path / subdir).exists()
        assert (tmp_path / subdir).is_dir()
        assert not (tmp_path / subdir).is_symlink()


def test_setup_directory_structure_noop(empty_run_dir_root: Path):
    cli_tools.setup_directory_structure(empty_run_dir_root)
    assert empty_run_dir_root.exists()
    for subdir in [paths.BEST_LINK, paths.LATEST_LINK, paths.PRODUCTION_RUN]:
        assert (empty_run_dir_root / subdir).exists()
        assert (empty_run_dir_root / subdir).is_dir()
        assert not (empty_run_dir_root / subdir).is_symlink()


def test_move_link_dir(run_dir_root: Path):
    link_dir = (run_dir_root / 'test_link_dir')
    link_dir.mkdir()
    target_dir = _get_random_run_dir(run_dir_root)
    cli_tools.move_link(link_dir, target_dir)
    assert link_dir.exists()
    assert link_dir.is_dir()
    assert link_dir.is_symlink()
    assert link_dir.resolve() == target_dir


def test_move_link_file_fail(run_dir_root: Path):
    link_file = (run_dir_root / 'test_link_dir')
    link_file.touch()
    target_dir = _get_random_run_dir(run_dir_root)
    with pytest.raises(ValueError, match='not a symlink or a directory'):
        cli_tools.move_link(link_file, target_dir)


def test_move_link_existing_link(run_dir_root: Path):
    link_dir = (run_dir_root / 'test_link_dir')
    link_dir.mkdir()
    target_dir = _get_random_run_dir(run_dir_root)
    cli_tools.move_link(link_dir, target_dir)
    assert link_dir.resolve() == target_dir
    new_target_dir = _get_random_run_dir(run_dir_root)
    cli_tools.move_link(link_dir, new_target_dir)
    assert link_dir.exists()
    assert link_dir.is_dir()
    assert link_dir.is_symlink()
    assert link_dir.resolve() == new_target_dir


def test_mark_explicit(run_dir_root: Path):
    run_dir = _get_random_run_dir(run_dir_root)
    for link_name in [paths.BEST_LINK, paths.LATEST_LINK, 'other_link']:
        cli_tools.mark_explicit(run_dir, run_dir_root, link_name)
        link_path = run_dir_root / link_name
        assert link_path.exists()
        assert link_path.is_dir()
        assert link_path.is_symlink()
        assert link_path.resolve() == run_dir


def test_mark_explicit_nested_version_root(run_dir_root: Path):
    run_dir = _get_random_run_dir(run_dir_root)
    prod_dir = run_dir_root / paths.PRODUCTION_RUN
    link_name = 'prod_run'
    link_path = prod_dir / link_name
    cli_tools.mark_explicit(run_dir, prod_dir, link_name)
    assert link_path.exists()
    assert link_path.is_dir()
    assert link_path.is_symlink()
    assert link_path.resolve() == run_dir


@pytest.mark.parametrize(('mark_func', 'link_name'),
                         ((cli_tools.mark_latest_explicit, paths.LATEST_LINK),
                          (cli_tools.mark_best_explicit, paths.BEST_LINK)),
                         ids=lambda x: x if isinstance(x, str) else '')
def test_mark_link_explicit(mark_func: Callable, link_name: str, run_dir_root: Path):
    run_dir = _get_random_run_dir(run_dir_root)
    prod_dir = run_dir_root / paths.PRODUCTION_RUN
    for root_dir in [run_dir_root, prod_dir]:
        link_path = root_dir / link_name
        mark_func(run_dir, root_dir)
        assert link_path.exists()
        assert link_path.is_dir()
        assert link_path.is_symlink()
        assert link_path.resolve() == run_dir


@pytest.mark.parametrize(('mark_func', 'link_name'),
                         ((cli_tools.mark_latest, paths.LATEST_LINK),
                          (cli_tools.mark_best, paths.BEST_LINK)),
                         ids=lambda x: x if isinstance(x, str) else '')
def test_mark_link(mark_func: Callable, link_name: str, run_dir_root: Path):
    run_dir = _get_random_run_dir(run_dir_root)
    mark_func(run_dir)
    link_path = run_dir_root / link_name
    assert link_path.exists()
    assert link_path.is_dir()
    assert link_path.is_symlink()
    assert link_path.resolve() == run_dir


def test_mark_production_explicit_no_date(run_dir_root: Path, mock_datetime):
    run_dir = _get_random_run_dir(run_dir_root)
    cli_tools.mark_production_explicit(run_dir, run_dir_root / paths.PRODUCTION_RUN)
    link_path = run_dir_root / paths.PRODUCTION_RUN / MOCK_DATETIME.strftime('%Y_%m_%d')
    assert link_path.exists()
    assert link_path.is_dir()
    assert link_path.is_symlink()
    assert link_path.resolve() == run_dir


def test_mark_production_explicit_fail(run_dir_root: Path):
    run_dir = _get_random_run_dir(run_dir_root)
    date = 'the_time_is_now'
    with pytest.raises(ValueError):
        cli_tools.mark_production_explicit(run_dir, run_dir_root / paths.PRODUCTION_RUN, date)


def test_mark_production_explicit(run_dir_root: Path):
    run_dir = _get_random_run_dir(run_dir_root)
    date = '2020_04_25'
    cli_tools.mark_production(run_dir, date)
    link_path = run_dir_root / paths.PRODUCTION_RUN / date
    assert link_path.exists()
    assert link_path.is_dir()
    assert link_path.is_symlink()
    assert link_path.resolve() == run_dir


def test_mark_production_no_date(run_dir_root: Path, mock_datetime):
    run_dir = _get_random_run_dir(run_dir_root)
    cli_tools.mark_production(run_dir)
    link_path = run_dir_root / paths.PRODUCTION_RUN / MOCK_DATETIME.strftime('%Y_%m_%d')
    assert link_path.exists()
    assert link_path.is_dir()
    assert link_path.is_symlink()
    assert link_path.resolve() == run_dir


def test_mark_production_fail(run_dir_root: Path):
    run_dir = _get_random_run_dir(run_dir_root)
    date = 'the_time_is_now'
    with pytest.raises(ValueError):
        cli_tools.mark_production(run_dir, date)


def test_mark_production(run_dir_root: Path):
    run_dir = _get_random_run_dir(run_dir_root)
    date = '2020_04_25'
    cli_tools.mark_production(run_dir, date)
    link_path = run_dir_root / paths.PRODUCTION_RUN / date
    assert link_path.exists()
    assert link_path.is_dir()
    assert link_path.is_symlink()
    assert link_path.resolve() == run_dir


@pytest.mark.parametrize(('last_stage_version', 'last_stage_directory', 'last_stage_root', 'result'),
                         [(ABSOLUTE_PATH_1, ABSOLUTE_PATH_2, None, ABSOLUTE_PATH_2),
                          (ABSOLUTE_PATH_1, ABSOLUTE_PATH_2, ABSOLUTE_PATH_1, ABSOLUTE_PATH_2),
                          (ABSOLUTE_PATH_1, None, None, ABSOLUTE_PATH_1),
                          (ABSOLUTE_PATH_1, None, ABSOLUTE_PATH_2, ABSOLUTE_PATH_1),
                          (None, ABSOLUTE_PATH_1, None, ABSOLUTE_PATH_1),
                          (ABSOLUTE_PATH_1, RELATIVE_PATH, ABSOLUTE_PATH_2, ABSOLUTE_PATH_2 / RELATIVE_PATH),
                          (RELATIVE_PATH, None, ABSOLUTE_PATH_1, ABSOLUTE_PATH_1 / RELATIVE_PATH),
                          ])
def test_get_last_stage_directory(last_stage_version, last_stage_directory, last_stage_root, result):
    assert cli_tools.get_last_stage_directory(last_stage_version, last_stage_directory, last_stage_root) == result


@pytest.mark.parametrize(('last_stage_version', 'last_stage_directory', 'last_stage_root'),
                         [(ABSOLUTE_PATH_1, RELATIVE_PATH, None),
                          (RELATIVE_PATH, None, None),
                          (RELATIVE_PATH, None, RELATIVE_PATH),
                          (RELATIVE_PATH, RELATIVE_PATH, RELATIVE_PATH)
                          ])
def test_get_last_stage_directory_errors(last_stage_version, last_stage_directory, last_stage_root):
    with pytest.raises(ValueError):
        cli_tools.get_last_stage_directory(last_stage_version, last_stage_directory, last_stage_root)



__all__ = [
    "__title__", "__summary__", "__uri__", "__author__", "__email__", "__version__",
]

__title__ = "covid-shared"
__summary__ = "Shared utilities for the covid model pipeline."
__uri__ = "https://stash.ihme.washington.edu/projects/CVD19/repos/covid-shared/browse"

__version__ = '1.0.50'

__author__ = "James Collins"
__email__ = "collijk@uw.edu"
